#include <errno.h>
#include <fcntl.h>
#include "menvars.h"
#include "menvers.h"
#include <grp.h>
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>
#include <wchar.h>


/*
*  This program reads data from a Menarini V-Sight serial port and dumps it
*  into a file.
*
*  Copyright (C) 2018 - 2021 R.J. van der Putten, Leiden, Holland, 
*  rob at sput dot nl.
* 
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License version 3 as
*  published by the Free Software Foundation.
* 
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*/


/*  TTY  */
int fd;
struct termios oldtio;

/*  Use pre in non-table HTML  */
//#define MENRD_USE_PRE 1

/*  Use timeout or pipe for SIGTERM  */
//#define MENRD_USE_TIMEOUT 1

#ifndef MENRD_USE_TIMEOUT
 /*  Pipe for select  */
 int pfd[2];
#endif

struct my_baud {
    unsigned int rate;
    char *str;
} my_bauds[] = { \
    {    B300,    "300" }, \
    {    B600,    "600" }, \
    {   B1200,   "1200" }, \
    {   B2400,   "2400" }, \
    {   B4800,   "4800" }, \
    {   B9600,   "9600" }, \
    {  B19200,  "19200" }, \
    {  B38400,  "38400" }, \
    {  B57600,  "57600" }, \
    { B115200, "115200" }, \
    {       0,    NULL  }  \
};

#define MENRD_DFLT_SPD "115200"

/*  Global stuff  */
int contprog;	/*  Stop program when 0  */

/*  Date - Time  */
char datstr[48];
struct timeval dstv;

/*  Logging  */
int   debfl;
char  logstr[512];
char  tologst[768];
pid_t pid;
FILE *logfil;

/*  Mail  */
char systr[768];

/*  String for gnuplot  */
char plotstr[4096];
/*  String for text output  */
char textstr[4096];


/*
    inspired by
    https://tldp.org/HOWTO/Serial-Programming-HOWTO/x115.html
*/


/*  Generate date string */
char * gendat(char *s, int mode)
{
    int len, ms;
    struct tm *time;

    if (s == NULL)
	return(NULL);

    /*
    *	Modes;
    *
    *	0  YYYYMMDD-hhmmss
    *	1  YYYY-MM-DD hh:mm:ss
    *	2  YYYYMMDD-hhmmss-ms
    *	3  YYYY-MM-DD hh:mm:ss.ms
    */

    gettimeofday(&dstv, NULL);
    ms = (int) ((long) dstv.tv_usec / 1000L);
    time = localtime(&dstv.tv_sec);

    if ((mode & 1) == 0) {
	len = snprintf(s, 20, "%04d%02d%02d-%02d%02d%02d", \
	  time->tm_year + 1900, time->tm_mon + 1, time->tm_mday, \
	  time->tm_hour, time->tm_min, time->tm_sec);
    } else {
	len = snprintf(s, 20, "%04d-%02d-%02d %02d:%02d:%02d", \
	  time->tm_year + 1900, time->tm_mon + 1, time->tm_mday, \
	  time->tm_hour, time->tm_min, time->tm_sec);
    }

    if ((mode & 2) != 0) {
	/*  Add milliseconds  */
	if ((mode & 1) == 0)
	    snprintf(s + len, 12, "-%03d", ms);
	else
	    snprintf(s + len, 12, ".%03d", ms);
    }

    return(s);
}

/*  Print to logfile or stderr  */
int prtolog(char *s1, char *s2)
{
    if(s1 == NULL)
        return(1);

    if (logfil == NULL) {
	/*  Interactive mode, print to stderr  */
	fprintf(stderr, "%s", s1);
	if (s2 != NULL)
	    fprintf(stderr, " %s", s2);
	fprintf(stderr, "\n");
	fflush(stderr);
    } else {
	/*  Deamon, print to log  */
	if(s2 == NULL)
	    snprintf(tologst, 767, "%s %s\n", gendat(datstr, 3), s1);
	else
	    snprintf(tologst, 767, "%s %s %s\n", gendat(datstr, 3), s1, s2);
	fprintf(logfil, "%s", tologst);
	fflush(logfil);
    }

    return(0);
}

/*  Stop program  */
void stop(int sig)
{
    #ifndef MENRD_USE_TIMEOUT
     char ch;
     ch = 'q';
    #endif

    contprog = 0;
    prtolog("stop(): recieved SIGTERM", NULL);
    #ifndef MENRD_USE_TIMEOUT
     /*  Write to pipe to make sure */
     if (write(pfd[1], &ch, 1) < 0) {
	 snprintf(logstr, 511, "stop(): Can't write to pipe: %s", \
	   strerror(errno));
	 prtolog(logstr, NULL);
     }
    #endif
}

/*  Reset TTY  */
void resetty(void)
{
    if (fd > 0) {
	/*  fd is TTY  */
	tcflush(fd, TCIOFLUSH);
	tcsetattr(fd, TCSANOW, &oldtio);
	close(fd);
    }
}

/*  Generate filename and open file  */
FILE * genfil(FILE *f1, char *filstr, char *path, char *prefix, int mode)
{
    /* f1 = genfil(f1, filstr, path, name); */

    if(path != NULL && path[0] != 0)
	snprintf(filstr, 367, "%s/%s-%s", path, prefix, gendat(datstr, 2));
    else 
	snprintf(filstr, 367, "%s-%s", prefix, gendat(datstr, 2));

    if (filstr[366] != 0) {
	prtolog("genfil(): Filename or path too long", NULL);
	return(NULL);
    }

    if ((mode & MENRD_HTML) != 0)
	strcat(filstr, ".html");
    else if ((mode & MENRD_RAW) != 0)
	strcat(filstr, ".raw");
    else
	strcat(filstr, ".txt");

    if ((f1 = fopen(filstr, "a")) == NULL) {
	snprintf(logstr, 511, "genfil(): Can't open %s: %s", \
	  filstr, strerror(errno));
	prtolog(logstr, NULL);
    }
 
    return(f1);
}

/*  Help  */
void hlp(void)
{
    fprintf(stderr, "Writes data from Menarini V-Sight to file.\n");
    fprintf(stderr, "Usage: menar-read [-a altpath] [-b baudrate] [-c] [-D] ");
    fprintf(stderr, "-d device | -f input file\n\t[-g gid] [-G] [-h] [-H] ");
    fprintf(stderr, "[-i] [-I imgpath] [-l logfile] [-m address]\n\t");
    fprintf(stderr, "[-n filename prexfix] [-N] [-p path] [-P e|n|o] ");
    fprintf(stderr, "[-r piddir] [-R] [-s]\n\t[-t] [-u uid] [-U] [-v] [-x]\n");
    fprintf(stderr, "-c: Enable RTS CTS flow control.\n");
    fprintf(stderr, "-D: Enable debug.\n");
    fprintf(stderr, "-G: Generate graphs.\n");
    fprintf(stderr, "-h: Print help and exit.\n");
    fprintf(stderr, "-H: HTML output.\n");
    fprintf(stderr, "-i: Ignore stdin.\n");
    fprintf(stderr, "-N: With normal values.\n");
    fprintf(stderr, "-P: Even, No or Odd parity. ");
    fprintf(stderr, "This also sets 7-bits.\n");
    fprintf(stderr, "-R: Raw output.\n");
    fprintf(stderr, "-s: Strip output.\n");
    fprintf(stderr, "-t: Use tables in HTML.\n");
    fprintf(stderr, "-U: UTF-8 output.\n");
    fprintf(stderr, "-v: Print version and exit.\n");
    fprintf(stderr, "-x: Enable XON XOFF flow control.\n");
    fprintf(stderr, "Default baudrate is %s bps.\n", MENRD_DFLT_SPD);
    fprintf(stderr, "Default filename is TTY-YYYYMMDD-hhmmss-ms.txt,\n");
    fprintf(stderr, "\tEG: ttyS1-20181224-135557-123.txt\n");
    fprintf(stderr, "The tty is set to RAW, 8N1. ");
    fprintf(stderr, "Or RAW, 7(E|N|O)1.\n");
    fprintf(stderr, "You need to set either device or input file.\n");
}

/*  Print version  */
void prtvers(void)
{
    printf("Version: %s\n", menvers);
    #ifdef MENRD_USE_TIMEOUT
     printf("Compiled with use timeout.\n");
    #endif
    #ifdef MENRD_USE_PRE
     printf("Compiled with use pre.\n");
    #endif
}

/* Convert baudrate into int  */
unsigned int findbaud(char *s)
{
    int i;

    i = 0;
    while (my_bauds[i].str != NULL) {
	if(strcmp(s, my_bauds[i].str) == 0)
	    return(my_bauds[i].rate);
	i++;
    }

    return(0);
}

/*  List baudrates  */
void listbauds(void)
{
    int i;

    fprintf(stderr, "%s", my_bauds[0].str);
    i = 1;
    while (my_bauds[i].str != NULL) {
	fprintf(stderr, ", %s", my_bauds[i].str);
	i++;
    }
}

/*  Mail errors  */
int mailerr(char *mes, char *adr)
{
    if(mes == NULL || adr == NULL)
	return(1);

    snprintf(systr, 767, "echo \"%s\" | mail -s \"Menar error\" \"%s\"", \
	mes, adr);
    system(systr);

    return(0);
}

/*
*  The stuff below is all about nice output
*/

/*  Is it a number?  */
int chkdigits(char *s, int len)
{
    int i;

    i = 0;
    while (i < len) {
	if (s[i] < '0' || s[i] > '9') {
	    return(-1);
	}
	i++;
    }

    return(0);
}

/* Fill array with graph values  */
int filgrvals(char *buf, int j, int *graphvals)
{
    char ch;
    int i;

    ch = 0;    
    i  = 0;

    while (i < 256) {
	if (chkdigits(buf + j, 3) == 0) {
	    /*  Is a number  */
	    ch = buf[j + 3];
	    buf[j + 3] = 0;
	    graphvals[i] = atoi(buf + j);
	    buf[j + 3] = ch;
	} else {
	    /*  Is not a number  */
	    graphvals[i] = 0;
	}
	i++;
	j += 3;
    }

    return(j);
}

/*  Plot the graph using gnuplot  */
int plotgraph(char *name, char *type, int *graphvals, int *graphlins, \
  int start, int cnt, char *color, char *title, int spec)
{
    int i, len, ret;
    FILE *pout;

    i = 0;
    len = 0;
    ret = 0;
    pout = NULL;

    /*  String for gnuplot  */
    ret = sprintf(plotstr + len, \
      "gnuplot -e \"set style line 2 lc rgb \\\"#%s\\\" lw 3 lt 1 ; ", \
      color);
    len += ret;
    ret = sprintf(plotstr + len, \
      "set terminal png size %d, %d ; ", MENRD_XSIZ, MENRD_YSIZ);
    len += ret;
    ret = sprintf(plotstr + len, \
      "set xrange [0:255] ; set yrange [0:255] ; ");
    len += ret;
    ret = sprintf(plotstr + len, \
      "unset ytics ; ");
    len += ret;
    ret = sprintf(plotstr + len, \
      "set output \\\"%s-%s.png\\\" ; ", name, type);
    len += ret;
    ret = sprintf(plotstr + len, \
      "set style data lines ; ");
    len += ret;

    /*  Tics  */
    if (strcmp(type, "wbc") == 0) {
	ret = sprintf(plotstr + len, "set xtics %s ; ", MENRD_WBC_TCS);
    } else if (strcmp(type, "rbc") == 0) {
	ret = sprintf(plotstr + len, "set xtics %s ; ", MENRD_RBC_TCS);
    } else {
	if (spec == MENRD_DOG) {
	    ret = sprintf(plotstr + len, "set xtics %s ; ", MENRD_PLT_TCS_DOG);
	} else if (spec == MENRD_CAT) {
	    ret = sprintf(plotstr + len, "set xtics %s ; ", MENRD_PLT_TCS_CAT);
	} else {
	    ret = sprintf(plotstr + len, "unset xtics ; ");
	}
    }
    len += ret;

    /*  Vertical lines  */
    for (i = 0; i < cnt; i++) {
	if (graphlins[start + i] > -1 && graphlins[start + i] < 256) {
	    ret = sprintf(plotstr + len, \
	      "set arrow %d from %d,0 to %d,255 nohead ; ", \
	      i + 1, graphlins[start + i], graphlins[start + i]);
	    len += ret;
	}
    }

    /*  The actual plot command  */
    ret = sprintf(plotstr + len, \
      "plot '-' using 1 ls 2 title \\\"%s\\\"\"", title);
    len += ret;

    /*  Open process  */
    if ((pout = popen(plotstr, "w")) == NULL) {
	fprintf(stderr, "Can't open gnuplot\n");
	exit(1);
    }

    /*  Print to gnuplot stdin  */
    i = 0;
    while (i < 256) {
	fprintf(pout, "%d\n", graphvals[i]);
	i++;
    }

    pclose(pout);

    return(0);
}

/*  Write the three graphs to file  */
int wrtgraphs(char *filstr, char *buf, int spec)
{
    char ch, name[512];
    int graphlins[8], graphvals[256], i, j, len;

    ch = 0;
    i  = 0;
    j  = 0;
    len = 0;
    memset(name, 0, 512);
    memset(graphlins, 0, sizeof(graphlins));
    memset(graphvals, 0, sizeof(graphvals));
    memset(plotstr, 0, sizeof(plotstr));

    strncpy(name, filstr, 511);
    len = strlen(name);
    if (strncmp(name + len - 4, ".raw", 4) == 0 || \
      strncmp(name + len - 4, ".txt", 4) == 0) {
	name[len - 4] = 0;
    } else if (strncmp(name + len - 5, ".html", 4) == 0) {
	name[len - 5] = 0;
    }

    /*
    *  Graphs. Order is wbc, rbc, plt
    */

    /*  Lines  */
    i = 0;
    j = MENRD_LINES_POS;
    while (i < 8) {
	if (chkdigits(buf + j, 3) == 0) {
	    ch = buf[j + 3];
	    buf[j + 3] = 0;
	    graphlins[i] = atoi(buf + j);
	    buf[j + 3] = ch;
	} else {
	    graphlins[i] = 0;
	}
	i++;
	j += 3;
    }

    /*  Graphs  */
    j = MENRD_GRAPH_POS;
    /*  White blood cell histogram  */
    j = filgrvals(buf, j, graphvals);
    plotgraph(name, "wbc", graphvals, graphlins, 0, 4, MENRD_COL_WBC, "WBC", \
      spec);
    /*  Red blood cell histogram  */
    j = filgrvals(buf, j, graphvals);
    plotgraph(name, "rbc", graphvals, graphlins, 4, 2, MENRD_COL_RBC, "RBC", \
      spec);
    /*  Platelet histogram  */
    j = filgrvals(buf, j, graphvals);
    plotgraph(name, "plt", graphvals, graphlins, 6, 2, MENRD_COL_PLT, "PLT", \
      spec);

    return(0);
}

/*  Generate **.** like string  */
char * repdigits(char *s, int len, int decml, int mode)
{
    int i, j, prefx, spaclen;

    i = 0;
    j = 0;

    /*  Prefix  */
    prefx = len - decml;
    if(prefx < 0)
	prefx = 0;	/*  Error  */

    /*  Space length  */
    spaclen = 4 - len + decml;
    if (spaclen > 3)
	spaclen = 3;	/*  0.dddd  */
    if (mode != 0) {
	/*  Prepend spaces  */
	while (j < spaclen) {
	    s[j] = ' ';
	    j++;
	}
    }

    if (prefx == 0) {
	/*  0.dddd  */
	s[j] = '0';
	j++;
	s[j] = '.';
	j++;
	while (i < len) {
	    s[j] = '*';
	    i++;
	    j++;
	}
	/*  Terminate  */
	s[j] = 0;
	return(s);
    }

    /*  dd.dd or dddd  */
    while (i < prefx) {
	s[j] = '*';
	i++;
	j++;
    }
    if (decml == 0) {
	/*  dddd  */
	s[j] = 0;
	return(s);
    }

    /*  dd.dd  */
    s[j] = '.';
    j++;
    while (i < len) {
	s[j] = '*';
	i++;
	j++;
    }
    s[j] = 0;

    return(s);
}

/*  Check blood value  */
int chkval(int spec, int valnum, double val)
{
    if(mnanlms[spec][valnum][0] != 0.0 && val < mnanlms[spec][valnum][0])
	return('L');
    else if(mnanlms[spec][valnum][1] != 0.0 && val > mnanlms[spec][valnum][1])
	return('H');

    return(' ');
}

/*  Print a number padded with spaces  */
int prtnum(char *s, int var, double val)
{
    if (menbldvars[var].decml == 0) {
	/*  No decimals  */
	return(sprintf(s, "%4d", (int) val));
    } else {
	return(sprintf(s, "%*.*f", 5 + menbldvars[var].decml, \
	  menbldvars[var].decml, val));
    }
}

/*  Write a text or HTML table  */
int wrtxt(FILE *f1, char *buf, int mode, int spec)
{
    char ch, day[4], hour[4], min[4], mon[4], repstr[16], year[6];
    int i, j, ret, stat, tot;
    double f;

    ch = 0;
    f  = 0.0;
    i  = 0;
    j  = 0;
    ret = 0;
    stat = 0;		/*  Blood value low, high or OK  */
    tot = 0;
    memset(day,    0, 4);
    memset(hour,   0, 4);
    memset(min,    0, 4);
    memset(mon,    0, 4);
    memset(repstr, 0, 16);
    memset(year,   0, 6);

    #ifdef MENRD_USE_PRE
     if ((mode & MENRD_HTML) != 0) {
	 if ((mode & MENRD_TABLE) == 0) {
	     ret = fprintf(f1, "<pre>\n");
	 } else {
	     /*  Fancy with tables  */
	     ret = fprintf(f1, \
	       "<h1 style=\"text-align: center;\">Results</h1>\n\n<p>\n");
	 }
	 if (ret < 1)
	     return(0);
	 tot += ret;
     }
    #else
     if ((mode & MENRD_TABLE) != 0) {
	 /*  Fancy with tables  */
	 ret = fprintf(f1, \
	   "<h1 style=\"text-align: center;\">Results</h1>\n\n<p>\n");
	 if (ret < 1)
	     return(0);
	 tot += ret;
     }
    #endif

    ch = buf[9];
    buf[9] = 0;
    if (datstr[0] == 0) {
	/*  Data from TTY; Generate date string  */
	gendat(datstr, 1);
    }
    if ((mode & MENRD_HTML) == 0 || (mode & MENRD_TABLE) == 0) {
	ret = fprintf(f1, "Saved to file at: %s\n\n", datstr);
	if (ret < 1)
	    return(0);
	tot += ret;
	ret = fprintf(f1, "Species: %s\n", menanimals[spec]);
	if (ret < 1)
	    return(0);
	tot += ret;
	ret = fprintf(f1, "Id: %s\n", buf + 1);
    } else {
	ret = fprintf(f1, "Saved to file at: %s<br>\n<br>\n", datstr);
	if (ret < 1)
	     return(0);
	tot += ret;
	ret = fprintf(f1, "Species: %s<br>\n", menanimals[spec]);
	if (ret < 1)
	    return(0);
	tot += ret;
	ret = fprintf(f1, "Id: %s<br>\n", buf + 1);
    }
    buf[9] = ch;
    if (ret < 1)
	return(0);
    tot += ret;

    if ((mode & MENRD_STRIP) == 0) {
	/*  Print date  */
	strncpy(year, buf + 14, 4);
	year[4] = 0;
	strncpy(mon,  buf + 10, 2);
	mon[2] = 0;
	strncpy(day,  buf + 12, 2);
	day[2] = 0;
	strncpy(hour, buf + 18, 2);
	hour[2] = 0;
	strncpy(min,  buf + 20, 2);
	min[2] = 0;

	if ((mode & MENRD_HTML) == 0 || (mode & MENRD_TABLE) == 0) {
	    ret = fprintf(f1, "Orig date: %s-%s-%s %2s:%2s\n", \
	      year, mon, day, hour, min);
	} else {
	    ret = fprintf(f1, "Orig date: %s-%s-%s %s:%s<br>\n", \
	      year, mon, day, hour, min);
	}
	if (ret < 1)
	    return(0);
	tot += ret;
    }

    if ((mode & MENRD_TABLE) == 0) {
	/*  Header  */
	ret = sprintf(textstr, "\nVar      Val        Unit");
	if ((mode & MENRD_NVALS) != 0) {
	    if ((mode & MENRD_UTF) == 0) {
		ret += sprintf(textstr + ret, "      Min       Max");
	    } else {
		ret += sprintf(textstr + ret , "     Min       Max");
	    }
	}
	/*  Horizontal line; Type and length set in main()  */
	ret += sprintf(textstr + ret, "\n%s\n", hlin);
    } else {
	ret = sprintf(textstr, \
	  "</p>\n\n<h2>Data</h2>\n\n<table>\n<thead>\n");
	ret += sprintf(textstr + ret, \
	  "<tr><th>Var</th><th>Val</th><th>H/L</th><th>Unit</th>");
	if ((mode & MENRD_NVALS) != 0) {
	    ret += sprintf(textstr + ret, \
	      "<th>Min</th><th>Max</th>");
	}
	strcat(textstr, "</tr>\n</thead>\n<tbody>\n");
    }
    ret = fprintf(f1, "%s", textstr);
    if (ret < 1)
	 return(0);
    tot += ret;

    i = 0;
    j = MENRD_DATA_POS;
    while (menbldvars[i].len != 0) {
	ch = buf[j + menbldvars[i].len];
	buf[j + menbldvars[i].len] = 0;
	if (menbldvars[i].mult != 0.0) {
	    /*  No padding  */
	    if (chkdigits(buf + j, menbldvars[i].len) == 0) {
		/*  It's a number  */
		f = atof(buf + j);
		f = f * menbldvars[i].mult;
		stat = chkval(spec, i, f);
		if ((mode & MENRD_TABLE) == 0) {
		    /*  No table  */
		    ret = sprintf(textstr, "%s:%s", \
		      menbldvars[i].name, mnrpds[menbldvars[i].pad]);
		    ret += prtnum(textstr + ret, i, f);
		    ret += sprintf(textstr + ret, "%s%c ", \
		      mnrpds[menbldvars[i].dcpad], stat);
		    if ((mode & MENRD_UTF) == 0) {
			ret += sprintf(textstr + ret, "%s", \
			  menbldvars[i].unita);
		    } else {
			ret += sprintf(textstr + ret, "%s", \
			  menbldvars[i].unitu);
		    }
		    if ((mode & MENRD_NVALS) != 0 && spec != MENRD_UNKNOWN && \
		     (mnanlms[spec][i][0] != 0.0 || \
		      mnanlms[spec][i][1] != 0.0)) {
			/*  Print standard values  */
			if ((mode & MENRD_UTF) == 0) {
			    ret += sprintf(textstr + ret, "%s", \
			      mnrpds[menbldvars[i].pada]);
			} else {
			    ret += sprintf(textstr + ret, "%s", \
			      mnrpds[menbldvars[i].padu]);
			}
			if (mnanlms[spec][i][0] == 0.0) {
			    ret += sprintf(textstr + ret, "%s", mnrpds[6]);
			} else {
			    ret += prtnum(textstr + ret, i, \
			      mnanlms[spec][i][0]);
			}
			if (mnanlms[spec][i][1] != 0.0) {
			    ret += sprintf(textstr + ret, "%s", \
			      mnrpds[menbldvars[i].dcpad]);
			    /* int prtnum(char *s, int var, double val) */
			    ret += prtnum(textstr + ret, i, \
			      mnanlms[spec][i][1]);
			}
		    }
		    strcat(textstr, "\n");
		    ret = fprintf(f1, "%s", textstr);
		} else {
		    /*  Use table  */
		    ret = sprintf(textstr, \
		      "<tr><td>%s:</td><td>%.*f</td><td>%c</td><td>%s</td>", \
		      menbldvars[i].name, menbldvars[i].decml, f, \
		      stat, menbldvars[i].unitu);
		    if ((mode & MENRD_NVALS) == 0) {
			strcat(textstr, "</tr>\n");
		    } else {
			if (spec == MENRD_UNKNOWN || \
			 (mnanlms[spec][i][0] == 0.0 && \
			  mnanlms[spec][i][1] == 0.0)) {
			    strcat(textstr, "<td> </td><td> </td></tr>\n");
			} else {
			    /*  Print standard values  */
			    if (mnanlms[spec][i][0] == 0.0) {
				ret += sprintf(textstr + ret, "<td> </td>");
			    } else {
				ret += sprintf(textstr + ret, \
				  "<td>%.*f</td>", \
				  menbldvars[i].decml, mnanlms[spec][i][0]);
			    }
			    if (mnanlms[spec][i][1] ==  0.0) {
				ret += sprintf(textstr + ret, \
				  "<td> </td></tr>\n");
			    } else {
				sprintf(textstr + ret, \
				  "<td>%.*f</td></tr>\n", \
				  menbldvars[i].decml, mnanlms[spec][i][1]);
			    }
			}
		    }
		    ret = fprintf(f1, "%s", textstr);
		}
	    } else {
		/*  Not a number! Something went wrong! */
		if ((mode & MENRD_UTF) == 0) {
		    /*  ASCII  */
		    ret = fprintf(f1, "%s:%s%s%s  %s\n", \
		      menbldvars[i].name, mnrpds[menbldvars[i].pad], \
		      repdigits(repstr, menbldvars[i].len, menbldvars[i].decml, 1), \
		      mnrpds[menbldvars[i].dcpad], menbldvars[i].unita);
		} else {
		    /*  UTF-8  */
		    if ((mode & MENRD_TABLE) == 0 || (mode & MENRD_HTML) == 0) {
			ret = fprintf(f1, "%s:%s%s%s  %s\n", \
			  menbldvars[i].name, mnrpds[menbldvars[i].pad], \
			  repdigits(repstr, menbldvars[i].len, menbldvars[i].decml, 1), \
			  mnrpds[menbldvars[i].dcpad], menbldvars[i].unitu);
		    } else {
			sprintf(textstr, \
			  "<tr><td>%s:</td><td>%s</td><td> </td><td>%s</td>", \
			  menbldvars[i].name, \
			  repdigits(repstr, menbldvars[i].len, menbldvars[i].decml, 0), \
			  menbldvars[i].unitu);
			if((mode & MENRD_NVALS) != 0)
			    strcat(textstr, "<td> </td><td> </td>");
			strcat(textstr, "</tr>\n");
			ret = fprintf(f1, "%s", textstr);
		    }
		}
	    }
	    if (ret < 1)
		return(0);
	    tot += ret;
	}	/*  End of if (menbldvars[i].mult != 0.0)  */
	buf[j + menbldvars[i].len] = ch;
	j = j + menbldvars[i].len;
	i++;
    }	/*  End of while (menbldvars[i].len != 0)  */

    if ((mode & MENRD_TABLE) == 0 || (mode & MENRD_HTML) == 0) {
	ret = fprintf(f1, "\n");
    } else {
	/*  End table  */
	ret = fprintf(f1, "</tbody>\n</table>\n\n");
    }
    if (ret < 1)
	 return(0);
    tot += ret;

    return(tot);
}

/*  Write raw to file  */
int wrtraw(FILE *f1, char *buf)
{
    int ret, tot;

    ret = fprintf(f1, "Saved to file at: %s\n\n", \
      gendat(datstr, 1));
    if(ret < 1)
	return(0);
    tot = ret;

    ret = fprintf(f1, "%s", buf);
    if(ret < 1)
	return(0);

    return(tot + ret);
}

/*  Write HTML to file  */
int wrthtm(FILE *f1, char *filstr, char *buf, int mode, int spec, char *imgpath)
{
    int i, len, ret, tot;
    char line[4096], name[512];

    i = 0;
    len = 0;
    ret = 0;
    tot = 0;
    memset(line, 0, 4096);
    memset(name, 0, 512);

    len = strlen(filstr);
    i = len;
    while (i > 0) {
	if (filstr[i] == '/') {
	    strcpy(name, filstr + i + 1);
	    break;
	}
	i--;
    }
    if(name[0] == 0)
	strcpy(name, filstr);
    len = strlen(name);
    if(strncmp(name + len - 5, ".html", 5) == 0)
	name[len - 5] = 0;
    len = 0;

    if ((mode & MENRD_TABLE) != 0) {
	strcpy(line, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\"\n");
	strcat(line, " \"http://www.w3.org/TR/html4/strict.dtd\">\n");
	strcat(line, "<html>\n<head>\n");
	strcat(line, "<meta http-equiv=\"Content-Type\" content=\"text/html; ");
	strcat(line, "charset=UTF-8\">\n<title>Results</title>\n</head>\n");
	strcat(line, "<body>\n\n");
	ret = fprintf(f1, "%s", line);
	if (ret < 1)
	    return(0);
	tot = ret;
    }

    /*  Tables  */
    ret = wrtxt(f1, buf, mode, spec);
    if (ret < 1)
	return(0);
    tot += ret;

    /* Graphs  */
    wrtgraphs(filstr, buf, spec);

    if ((mode & MENRD_TABLE) != 0) {
	ret = fprintf(f1, "<h2>Graphs</h2>\n\n<p>\n");
	if (ret < 1)
	    return(0);
	tot += ret;
    }

    if (imgpath[0] == 0) {
	ret = fprintf(f1, \
	  "<img src=\"%s-wbc.png\" alt=\"White Blood Cell histogram\">\n", \
	  name);
    } else {
	ret = fprintf(f1, \
	  "<img src=\"%s/%s-wbc.png\" alt=\"White Blood Cell histogram\">\n", \
	  imgpath, name);
    }
    if (ret < 1)
	return(0);
    tot += ret;
    if (imgpath[0] == 0) {
	ret = fprintf(f1, \
	  "<img src=\"%s-rbc.png\" alt=\"Haemoglobin histogram\">\n", name);
    } else {
	ret = fprintf(f1, \
	  "<img src=\"%s/%s-rbc.png\" alt=\"Haemoglobin histogram\">\n", \
	  imgpath, name);
    }
    if (ret < 1)
	return(0);
    tot += ret;
    if (imgpath[0] == 0) {
	ret = fprintf(f1, \
	  "<img src=\"%s-plt.png\" alt=\"Platelet histogram\">\n", name);
    } else {
	ret = fprintf(f1, \
	  "<img src=\"%s/%s-plt.png\" alt=\"Platelet histogram\">\n", \
	  imgpath, name);
    }
    if (ret < 1)
	return(0);
    tot += ret;

    #ifdef MENRD_USE_PRE
     if ((mode & MENRD_TABLE) == 0) {
	  ret = fprintf(f1, "</pre>\n");
     } else {
	 ret = fprintf(f1, "</p>\n\n</body>\n</html>\n");
     }
     if (ret < 1)
	 return(0);
     tot += ret;
    #else
     if ((mode & MENRD_TABLE) != 0) {
	 ret = fprintf(f1, "</p>\n\n</body>\n</html>\n");
	 if (ret < 1)
	     return(0);
	 tot += ret;
     }
    #endif

    return(tot);
}

/*  Process output buffer  */
int procout(FILE *f1, char *filstr, char *buf, int len, int mode, char *imgpath)
{
    char ch;
    int j, flen, ret, spec;

    ch   = 0;
    j    = 0;
    flen = 0;
    ret  = 0;
    spec = 0;	/*  Species. Defaults to dog  */

    /*  Find species  */
    j = MENRD_SPECIES_POS;
    if (chkdigits(buf + j, 3) == 0) {
	/*  Is number  */
	ch = buf[j + 3];
	buf[j + 3] = 0;
	spec = atoi(buf + j);
	buf[j + 3] = ch;
	if (spec < 0 || spec > MENRD_ANIMAL3) {
	    spec = MENRD_UNKNOWN;
	    prtolog("procout(): Unknown species", NULL);
	}
    } else {
	/*  Is not a number  */
	spec = MENRD_UNKNOWN;
	prtolog("procout(): Unknown species", NULL);
    }

    flen = strlen(filstr);
    if ((mode & MENRD_RAW) != 0 || \
      strncmp(filstr + flen - 4, ".raw", 4) == 0 || \
      buf[0] != 'A' || len < MENRD_MIN_SIZ) {
	/*  Raw  */
	ret = wrtraw(f1, buf);
    } else if ((mode & MENRD_HTML) != 0) {
	/*  HTML  */
	ret = wrthtm(f1, filstr, buf, mode, spec, imgpath);
    } else {
	/*  Text  */
	ret = wrtxt(f1, buf, mode, spec);
    }

    if (ret < 1) {
	prtolog("procout(): Can't write to file", NULL);
	resetty();
	if(logfil != NULL)
	    fclose(logfil);
	fclose(f1);
	exit(1);
    }

    if ((mode & MENRD_GRAPH) != 0 && \
      (mode & MENRD_HTML) == 0) {
	/*  Write graphs  */
	wrtgraphs(filstr, buf, spec);
    }

    return(ret);
}

/*  Process raw input file  */
int procfil(FILE *fout, char *filstr, char *line, int outmod, char *imgpath)
{
    int len, tot;
    FILE *fin;

    len = 0;
    tot = 0;
    fin = NULL;

    /*  Input file  */
    len = strlen(filstr);
    if (len < 5 || strncmp(filstr + len - 4, ".raw", 4) != 0) {
	fprintf(stderr, "Must specify raw file\n");
	exit(1);
    }
    if ((fin = fopen(filstr, "r")) == NULL) {
	fprintf(stderr, "Can't open %s\n", filstr);
	exit(1);
    }

    /*  Output file  */
    filstr[len - 4] = 0;
    if ((outmod & MENRD_HTML) == 0) {
	strcat(filstr, ".txt");
	filstr[len] = 0;
    } else {
	strcat(filstr, ".html");
	filstr[len + 1] = 0;
    }
    /*  filstr is now ouput file name  */
    if ((fout = fopen(filstr, "w")) == NULL) {
	fprintf(stderr, "Can't open %s\n", filstr);
	fclose(fin);
	exit(1);
    }

    while (fgets(line, 4096, fin)) {
	len = strlen(line) - 1;
	if (line[len] == 10) {
	    line[len] = 0;
	    len--;
	}
	if (line[len] == 13) {
	    line[len] = 0;
	    len--;
	}
	len++;
	if (strncasecmp(line, "Saved to file at: ", 18) == 0) {
	    /*  Copy date from raw file  */
	    strncpy(datstr, line + 18, 47);
	}
	if (line[0] == 'A') {
	    break;
	}
    }
    if (line[0] != 'A') {
	fprintf(stderr, "Data not found\n");
	fclose(fout);
	fclose(fin);
	exit(1);
    }
    if (datstr[0] == 0) {
	/*  Raw file without date  */
	strcpy(datstr, "Unknown");
    }
    tot = procout(fout, filstr, line, len, outmod, imgpath);
    if (debfl != 0) {
	fprintf(stderr, "Debug: Wrote %d bytes to procout()\n", len);
	fprintf(stderr, "Debug: Wrote %d bytes to file %s\n", tot, filstr);
    }

    fclose(fout);
    fclose(fin);
    fout = NULL;

    return(tot);
}

/*  Read config file  */
int readconf(int mode)
{
    char line[4096], var[16];
    int i, j, len, spec;
    double min, max;
    FILE *f1;

    i = 0;
    j = 0;
    len = 0;
    max = 0.0;
    min = 0.0;
    spec = 0;
    memset(line, 0, 4096);
    memset(var, 0, 16);
    f1 = NULL;

    if ((f1 = fopen(MENRD_CONF, "r")) == NULL) {
	prtolog("Can't open file:", MENRD_CONF);
	return(1);
    }

    while (fgets(line, 4096, f1)) {
	if(line[0] == 10 || line[0] == 13 || line[0] == '#')
	    continue;
	len = strlen(line);
	if(len < 5)
	    continue;
	if (line[0] == '[') {
	    /*  Species  */
	    i = 1;
	    while (i < len) {
		if (line[i] == ']') {
		    line[i] = 0;
		    len = i + 1;
		    break;
		}
		i++;
	    }
	    j = 0;
	    while (j < MENRD_UNKNOWN) {
		if(strcasecmp(line + 1, menanimals[j]) == 0)
		    break;
		j++;
	    }
	    if(j >= MENRD_UNKNOWN)
		continue;
	    /*  Found species  */
	    spec = j;
	} else {
	    if (sscanf(line, "%15s %lf %lf", var, &min, &max) != 3)
		continue;
	    j = 0;
	    while (j < MENRD_BLD_VARS) {
		if (menbldvars[j].mult != 0.0 && \
		  strcasecmp(var, menbldvars[j].name) == 0) {
		    /*  Found variable  */
		    mnanlms[spec][j][0] = min;
		    mnanlms[spec][j][1] = max;
		    break;
		}
		j++;
	    }
	    if (debfl != 0 && j < MENRD_BLD_VARS) {
		if ((mode & MENRD_UTF) == 0) {
		    sprintf(logstr, "%s\t%s\t%7.3f\t%7.3f\t%s", \
		      menanimals[spec], menbldvars[j].name, min, max, \
		      menbldvars[j].unita);
		} else {
		    sprintf(logstr, "%s\t%s\t%7.3f\t%7.3f\t%s", \
		      menanimals[spec], menbldvars[j].name, min, max, \
		      menbldvars[j].unitu);
		}
		prtolog("Debug: readconf():", logstr);
	    }
	}
    }

    return(0);
}


int main(int argc, char **argv)
{
    int ch, i, len, maxfd, op, opt, par, retcnt, retval, tot;
    int flowfl, iginfl, xonfl;
    int outmod;
    unsigned int baudrate;
    char altpath[256], baudstr[16], buf[4096], dev[256], filstr[384];
    char imgpath[256], lgflstr[256], mailstr[64], mestr[512], name[64];
    char obuf[4096], path[256], pidf[384], pidir[256];
    char sendch;
    uid_t uid;
    gid_t gid;
    struct group *group;
    struct passwd *passwd;
    struct termios newtio;
    #ifdef MENRD_USE_TIMEOUT
     struct timeval tv;
    #endif
    FILE *f1;
    fd_set readfs;

    ch     = 0;
    debfl  = 0;		/*  Debug on when 1  */
    fd     = 0;		/*  File Descriptor for TTY  */
    flowfl = 0;		/*  CRTSCTS when 1  */
    gid    = 0;		/*  Group Id  */
    i      = 0;
    iginfl = 0;		/*  Ignore stdin when 1  */
    len    = 0;		/*  Length  */
    maxfd  = 0;
    op     = 0;		/*  Output pointer  */
    opt    = 0;
    outmod = 0;		/*  Output mode  */
    par    = 0;
    pid    = 0;		/*  Process Id  */
    retcnt = 0;		/*  Retry count  */
    retval = 0;
    sendch = 0;		/*  Char to be send to Tty  */
    tot    = 0;		/*  Total bytes written  */
    uid    = 0;		/*  User Id  */
    xonfl  = 0;		/*  XON XOFF when 1  */
    memset(altpath, 0, sizeof(altpath));	/*  Alternative path  */
    memset(buf,     0, sizeof(buf));
    memset(datstr,  0, sizeof(datstr));		/*  Date time string  */
    memset(dev,     0, sizeof(dev));		/*  Tty  */
    memset(&dstv,   0, sizeof(dstv));
    memset(filstr,  0, sizeof(filstr));		/*  Output file name  */
    memset(imgpath, 0, sizeof(imgpath));	/*  Img path in HTML  */
    memset(lgflstr, 0, sizeof(lgflstr));	/*  Logfile name  */
    memset(logstr,  0, sizeof(logstr));		/*  String for logfile  */
    memset(mailstr, 0, sizeof(mailstr));	/*  Email address of admin  */
    memset(mnanlms, 0, sizeof(mnanlms));	/*  Minum and maximum blood values  */
    memset(mestr,   0, sizeof(mestr));		/*  String for email message  */
    memset(name,    0, sizeof(name));		/*  File name prefix  */
    memset(&newtio, 0, sizeof(newtio));		/*  This makes the tty rather raw  */
    memset(obuf,    0, sizeof(obuf));		/*  Ouput buffer  */
    memset(path,    0, sizeof(path));		/*  Path to output files  */
    #ifndef MENRD_USE_TIMEOUT
     memset(pfd,     0, sizeof(pfd));		/*  Pipe Fd  */
    #endif
    memset(pidf,    0, sizeof(pidf));		/*  Pid file  */
    memset(pidir,   0, sizeof(pidir));		/*  Pid directory  */
    memset(systr,   0, sizeof(systr));		/*  String for system()  */
    memset(textstr, 0, sizeof(textstr));	/*  String for text output  */
    memset(tologst, 0, sizeof(tologst));
    #ifdef MENRD_USE_TIMEOUT
     memset(&tv,     0, sizeof(tv));
    #endif
    f1     = NULL;
    group  = NULL;
    logfil = NULL;
    passwd = NULL;

    /*  Default speed  */
    strncpy(baudstr, MENRD_DFLT_SPD, 15);
    /*  Continue program  */
    contprog = 1;
    /*  Get the pid  */
    pid = getpid();

    /*  Command line processing  */
    while ((opt = getopt(argc, argv, \
      "a:b:cd:Df:g:GhHiI:l:m:n:Np:P:r:Rstu:Uvx")) != -1) {
	switch (opt) {
	    case 'a':
		strncpy(altpath, optarg, 255);
		break;
	    case 'b':
		strncpy(baudstr, optarg, 15);	/*  < 255  */
		break;
	    case 'c':
		/*  Use RTS CTS flow control  */
		flowfl = 1;
		break;
	    case 'd':
		strncpy(dev, optarg, 255);
		break;
	    case 'D':
		debfl = 1;
		break;
	    case 'f':
		strncpy(filstr, optarg, 255);
		break;
	    case 'g':
		gid = (gid_t) atoi(optarg);
		if (gid < 1 || gid > 65535) {
		    /*  Group not numeric or invalid  */
		    group = getgrnam(optarg);
		    if (group == NULL) {
			fprintf(stderr, "Group not found\n");
			exit(1);
		    } else {
			gid = group->gr_gid;
		    }
		}
		break;
	    case 'G':
		/*  Output graphs  */
		outmod |= MENRD_GRAPH;
		break;
	    case 'h':
		hlp();
		exit(0);
	    case 'H':
		/*  HTML output  */
		outmod |= MENRD_HTML;
		break;
	    case 'i':
		/*  Ignore stdin  */
		iginfl = 1;
		break;
	    case 'I':
		/*  IMG path in HTML  */
		strncpy(imgpath, optarg, 255);
		break;
	    case 'l':
		strncpy(lgflstr, optarg, 255);
		break;
	    case 'm':
		strncpy(mailstr, optarg, 63);	/*  < 255  */
		break;
	    case 'n':
		strncpy(name, optarg, 63);	/*  < 255  */
		break;
	    case 'N':
		/*  Normal values in output  */
		outmod |= MENRD_NVALS;
		break;
	    case 'p':
		strncpy(path, optarg, 255);
		break;
	    case 'P':
		if (strcmp(optarg, "e") == 0) {
		    par = 'E';
		} else if (strcmp(optarg, "n") == 0) {
		    par = 'N';
		} else if (strcmp(optarg, "o") == 0) {
		    par = 'O';
		} else {
		    fprintf(stderr, "Parity should be 'e', 'n' or 'o'\n");
		    exit(1);
		}
		break;
	    case 'r':
		strncpy(pidir, optarg, 255);
		break;
	    case 'R':
		/*  Raw output  */
		outmod |= MENRD_RAW;
		break;
	    case 's':
		/*  Strip output  */
		outmod |= MENRD_STRIP;
		break;
	    case 't':
		/*  With tables  */
		outmod |= MENRD_TABLE;
		break;
	    case 'u':
		uid = (uid_t) atoi(optarg);
		if (uid < 1 || uid > 65535) {
		    /*  User not numeric or invalid  */
		    passwd = getpwnam(optarg);
		    if (passwd == NULL) {
			fprintf(stderr, "User not found\n");
			exit(1);
		    } else {
			uid = passwd->pw_uid;
		    }
		}
		break;
	    case 'U':
		/*  UTF-8 output  */
		outmod |= MENRD_UTF;
		break;
	    case 'v':
		prtvers();
		exit(0);
	    case 'x':
		xonfl = 1;
		break;
	    default:
		hlp();
		exit(1);
	}
    }

    /*  Dependencies  */
    if ((outmod & MENRD_HTML) != 0) {
	/*  HTML: graphs, UTF-8 and not raw  */
	outmod |= MENRD_GRAPH;
	outmod &= (~MENRD_RAW);
	outmod |= MENRD_UTF;
    } else if ((outmod & MENRD_RAW) != 0) {
	/*  Raw: no graphs, no norm values, no strip, no UTF-8  */
	outmod &= (~MENRD_GRAPH);
	outmod &= (~MENRD_NVALS);
	outmod &= (~MENRD_STRIP);
	outmod &= (~MENRD_UTF);
    }
    /*  No HTML implies no tables  */
    if ((outmod & MENRD_HTML) == 0)
	outmod &= (~MENRD_TABLE);
    /*  No graphs implies no img path  */
    if ((outmod & MENRD_GRAPH) == 0)
	imgpath[0] = 0;

    /* Set horizontal line length and type and length if any  */
    if ((outmod & MENRD_TABLE) == 0) {
	/*  No table  */
	if ((outmod & MENRD_NVALS) == 0) {
	    /*  No min and max values  */
	    if ((outmod & MENRD_UTF) == 0) {
		for(i = 0; i < MENRD_HL_A; i++)
		    strcat(hlin, "-");
	    } else {
		/*  U2500, BOX DRAWINGS LIGHT HORIZONTAL  */
		for(i = 0; i < MENRD_HL_U; i++)
		    strcat(hlin, "\xe2\x94\x80");
	    }
	} else {
	    /*  With min and max values  */
	    if ((outmod & MENRD_UTF) == 0) {
		for(i = 0; i < MENRD_HL_NV_A; i++)
		    strcat(hlin, "-");
	    } else {
		/*  U2500, BOX DRAWINGS LIGHT HORIZONTAL  */
		for(i = 0; i < MENRD_HL_NV_U; i++)
		    strcat(hlin, "\xe2\x94\x80");
	    }
	}
	i = 0;
    }

    if (filstr[0] != 0) {
	/* Implies no raw output  */
	outmod &= (~MENRD_RAW);
	/*  Read config  */
	readconf(outmod);
	/*  Convert raw input file and exit  */
	procfil(f1, filstr, buf, outmod, imgpath);
	contprog = 0;
	exit(0);
    }

    /*  Open logfile if specified  */
    if (lgflstr[0] != 0) {
	if((logfil = fopen(lgflstr, "a")) == NULL)
	    fprintf(stderr, "Can't open logfile %s: %s\n", 
	      lgflstr, strerror(errno));
    }

    /*  Read config  */
    readconf(outmod);

    if ((baudrate = findbaud(baudstr)) == 0) { 
	prtolog("Invalid baudrate:", baudstr);
	if (logfil == NULL) {
	    /*  List in interactive mode only  */
	    fprintf(stderr, "baudrate must be one of: ");
	    listbauds();
	    fprintf(stderr, ".\n");
	} else {
	    fclose(logfil);
	}
	exit(1);
    }
    if (dev[0] == 0) {
	prtolog("Must specify input file or device. EG: -d /dev/ttyS1", NULL);
	if(logfil != NULL)
	    fclose(logfil);
	exit(1);
    }

    /*  File name. If no prefix is specified, use tty name instead  */
    if (name[0] == 0) {
	/*  Strip path from tty  */
	i = strlen(dev) - 1;	/*  Start op last char  */
        while (i > 0) {
	    if (dev[i] == '/') {
		strncpy(name, dev + i + 1, 63);
		break;
	    }
	    i--;
	}
    } else {
	len = strlen(name) - 1;
	if (name[len] == '/') {
	    /*  Strip strailing slash  */
	    name[len] = 0;
	}
    }
    if (name[0] == 0) {
	prtolog("Could not set filename.", NULL);
	if(logfil != NULL)
	    fclose(logfil);
	exit(1);
    }

    /*  Path  */
    len = strlen(path) - 1;
    if (path[len] == '/') {
	/*  Strip strailing slash  */
	path[len] = 0;
    }
    /*  Alternative path  */
    if (altpath[0] != 0) {
	len = strlen(altpath) - 1;
	if(altpath[len] == '/')
	    altpath[len] = 0;
    }
    /*  Img path  */
    if (imgpath[0] != 0) {
	len = strlen(imgpath) - 1;
	if(imgpath[len] == '/')
	    imgpath[len] = 0;
    }

    #ifndef MENRD_USE_TIMEOUT
     /*  Pipe to assist quit on SIGTERM  */
     if (pipe(pfd) < 0) {
	snprintf(logstr, 511, "Can't open pipe: %s", strerror(errno));
	prtolog(logstr, NULL);
	if(logfil != NULL)
	    fclose(logfil);
	exit(1);
     }
    #endif

    /*
    *	Open modem device for reading and writing and not as controlling tty
    *	because we don't want to get killed if line noise sends CTRL-C.
    */
    if ((fd = open(dev, O_RDWR | O_NOCTTY | O_NONBLOCK)) < 0) {
	snprintf(logstr, 511, "Can't open device %s: %s", dev, strerror(errno));
	prtolog(logstr, NULL);
	if(logfil != NULL)
	    fclose(logfil);
	exit(1);
    }

    /*  Pidfile name  */
    if (pidir[0] == 0) {
	snprintf(pidf, 383, "/run/%s.pid", name);
    } else {
	len = strlen(pidir) - 1;
	if (pidir[len] == '/') {
	    /*  Strip strailing slash  */
	    pidir[len] = 0;
	}
	snprintf(pidf, 383, "%s/%s.pid", pidir, name);
    }

    /*  UID and GID  */
    if (uid != 0) {
	if (gid == 0)
	    gid = uid;
	if (setgid(gid) < 0) {
	    snprintf(logstr, 511, "Can't set GID to %d: %s", \
	      gid, strerror(errno));
	    prtolog(logstr, NULL);
	    gid = 0;
	}
	if (setuid(uid) < 0) {
	    snprintf(logstr, 511, "Can't set UID to %d: %s", \
	      uid, strerror(errno));
	    prtolog(logstr, NULL);
	    uid = 0;
	}
    }

    /*  Open pid file as new user  */
    if ((f1 = fopen(pidf, "w")) == NULL) {
	snprintf(logstr, 511, "Can't open pid file %s: %s", \
	  pidf, strerror(errno));
	prtolog(logstr, NULL);
	pidf[0] = 0;
    } else {
	fprintf(f1, "%d\n", pid);
	fclose(f1);
    }
    len = 0;
    f1 = NULL;

    /*  save current serial port settings  */
    tcgetattr(fd, &oldtio);

    /*
    BAUDRATE: Set bps rate. You could also use cfsetispeed and cfsetospeed.
    CRTSCTS:  Output hardware flow control.
    CLOCAL:   Ignore modem control (DSR, DCD) lines.
    CREAD:    Enable receiving characters.
    CS8:      8 bit, no parity, 1 stopbit.
    CS7:      7 bit with optional parity
    CREAD:    enable receiving characters
    HUPCL:    Lower modem control lines (DTR) when done.
    IXON:     Enable software flow control (outgoing)
    IXOFF:    Enable software flow control (incoming)
    */
    newtio.c_cflag = baudrate | HUPCL | CLOCAL | CREAD;
    /*  Flow control  */
    if (par == 0) {
	/*  Eight bits  */
	newtio.c_cflag = newtio.c_cflag | CS8;
	/*  Input: IGNPAR: ignore bytes with parity errors  */
	newtio.c_iflag = IGNPAR;
    } else {
	/*  Seven bits  */
	newtio.c_cflag = newtio.c_cflag | CS7;
	if (par != 'N') {
	    /*  Enable parity  */
	    newtio.c_cflag = newtio.c_cflag | PARENB;
	    if (par == 'O') {
		/*  Odd parity  */
		newtio.c_cflag = newtio.c_cflag | PARODD;
	    }
	    /*  Parity checking on input and strip high bit  */
	    newtio.c_iflag = newtio.c_iflag | INPCK | ISTRIP;
	}
    }
    if (flowfl != 0)
	newtio.c_cflag = newtio.c_cflag | CRTSCTS;
    if (xonfl  != 0)
	newtio.c_cflag = newtio.c_cflag | IXON | IXOFF;
    /*
    *	Blocking read until 1 character arrives.
    *	You don't need this with a non blocking read.
    */
    /* newtio.c_cc[VMIN] = 1; */

    /*  Clean the modem line and activate the settings for the port  */
    tcflush(fd, TCIOFLUSH);
    if (tcsetattr(fd, TCSANOW, &newtio) != 0) {
	snprintf(logstr, 511, "Can't set %s attributes: %s", dev, strerror(errno));
	prtolog(logstr, NULL);
	resetty();
	if(logfil != NULL)
	    fclose(logfil);
	exit(1);
    }

    /*  Print info  */
    prtolog("Menar-read version:", menvers);
    #ifdef MENRD_USE_TIMEOUT
     prtolog("Compiled with use timeout", NULL);
    #endif
    snprintf(logstr, 511, "Device %s, speed %s bps", dev, baudstr);
    prtolog(logstr, NULL);
    if (par == 0) {
	prtolog("Eight bits", NULL);
    } else if (par == 'E') {
	prtolog("Seven bits, Even parity", NULL);
    } else if (par == 'O') {
	prtolog("Seven bits, Odd parity", NULL);
    } else {
	prtolog("Seven bits, No parity", NULL);
    }
    if(flowfl != 0)
	prtolog("RTS CTS flow control", NULL);
    if(xonfl  != 0)
	prtolog("XON XOFF flow control", NULL);
    prtolog("Prefix:", name);
    if(path[0] != 0)
	prtolog("Files saved in:", path);
    if(altpath[0] != 0)
	prtolog("Alt path:", altpath);
    if ((outmod & MENRD_HTML) != 0)
	prtolog("HTML output", NULL);
    if ((outmod & MENRD_TABLE) != 0)
	prtolog("With tables", NULL);
    if ((outmod & MENRD_NVALS) != 0)
	prtolog("With standard values", NULL);
    if ((outmod & MENRD_GRAPH) != 0)
	prtolog("With graphs", NULL);
    if (imgpath[0] != 0)
	prtolog("With img path:", imgpath);
    if ((outmod & MENRD_RAW) != 0)
	prtolog("Raw output", NULL);
    if ((outmod & MENRD_UTF) != 0)
	prtolog("UTF-8 output", NULL);

    if(pidf[0] != 0)
	prtolog("Pid file:", pidf);
    if(uid != 0 || gid != 0) {
	snprintf(logstr, 511, "Uid: %d, Gid: %d", uid, gid);
	prtolog(logstr, NULL);
    }
    if (iginfl == 0 && logfil == NULL)
	printf("Press 'q' <Enter> to quit.\n");

    /*  Catch sigals  */
    //signal(SIGHUP,  stop);
    signal(SIGTERM,  stop);

    while (contprog != 0) {
	if (sendch != 0) {
	    /*  Write stuff before select  */
	    retcnt = 0;
	    if (write(fd, &sendch, 1) < 1) {
		if (errno == EAGAIN || errno == EINTR) {
		    prtolog("Retrying write to serial port", NULL);
		} else {
		    snprintf(logstr, 511, "Can't write to tty: %s", \
		      strerror(errno));
		    resetty();
		    fclose(f1);
		    if(logfil != NULL)
			fclose(logfil);
		    exit(1);
		}
		retcnt++;
		if (retcnt > 32 ) {
		    prtolog("Too many retries writing to serial port", NULL);
		    resetty();
		    fclose(f1);
		    if(logfil != NULL)
			fclose(logfil);
		    exit(1);
		}
		if (debfl != 0)
		    prtolog("Debug: Wrote 1 byte to serial port", NULL);
	    }
	    sendch = 0;	/*  Done  */
	}

	/*  XXX SIGTERM Test XXX  */
	//fprintf(stderr, "Input here\n");
	//read(0, &ch, 1);

	/*  (Re) Initialise select()  */
	FD_ZERO(&readfs);
	if (iginfl == 0) {
	    /*  Do not ignore stdin  */
	    FD_SET(0, &readfs);
	}
	FD_SET(fd, &readfs);
	#ifdef MENRD_USE_TIMEOUT
	 /*  Set the timeout to 1 s  */
	 tv.tv_sec  = 1;
	 tv.tv_usec = 0;
	 maxfd = fd + 1;
	 retval = select(maxfd, &readfs, NULL, NULL, &tv);
	#else
	 FD_SET(pfd[0], &readfs);
	 if (pfd[0] > fd)
	     maxfd = pfd[0] + 1;
	 else
	     maxfd = fd + 1;
	 retval = select(maxfd, &readfs, NULL, NULL, NULL);
	#endif
	if (retval < 0) {
	    if(contprog == 0)
		break;
	    if (errno == EINTR) {
		prtolog("Reinitilizing select()", NULL);
		continue;
	    } else {
		snprintf(logstr, 511, "Select failed: %s", strerror(errno));
		prtolog(logstr, NULL);
		resetty();
		fclose(f1);
		if(logfil != NULL)
		    fclose(logfil);
		exit(1);
	    }
	}

	/*  Block until read  */
	if (iginfl == 0 && FD_ISSET(0, &readfs)) {
	    /*  Check continue first  */
	    if(contprog == 0)
		break;
	    /*  Read data from keyboard  */
	    read(0, &ch, 1);
	    if (ch == 'Q' || ch == 'q') {
		prtolog("Quitting", NULL);
		contprog = 0;	/*  Don't continue  */
	    }
	}
	#ifndef MENRD_USE_TIMEOUT
	 if (FD_ISSET(pfd[0], &readfs)) {
	     if(contprog == 0)
		 break;
	     /*  Read data from pipe  */
	     read(pfd[0], &ch, 1);
	     contprog = 0;
	 }
	#endif
	if (FD_ISSET(fd, &readfs)) {
	    /*  Check continue first  */
	    if(contprog == 0)
		break;
	    /*  Read data from ttySx  */
	    len = read(fd, buf, 4095);
	    if (len < 1) {
		if(contprog == 0)
		    break;
		if (errno == EAGAIN || errno == EINTR) {
		    prtolog("Reinitilizing select()", NULL);
		    continue;
		} else {
		    snprintf(logstr, 511, "Can't read from tty: %s", \
		      strerror(errno));
		    prtolog(logstr, NULL);
		    resetty();
		    fclose(f1);
		    if(logfil != NULL)
			fclose(logfil);
		    exit(1);
		}
	    }
	    buf[len] = 0;	/*  Terminate  */
	    if (debfl != 0) {
		snprintf(logstr, 511, \
		  "Debug: Received %d bytes from serial port", len);
		prtolog(logstr, NULL);
	    }
	    if (debfl != 0) {
		write(1, buf, len);
	    }
	    i = 0;
	    sendch = 0;
	    while (i < len) {
		/*  Check comm prot per byte  */
		if (buf[i] == 0x02 || buf[i] == 0x05) {
		    /*  STX or ENQ  */
		    if (buf[i] == 0x05) {
			/*  ENQ, send ACK  */
			sendch = 0x06;
		    }
		    if (f1 == NULL) {
			/*  Open file  */
			if(debfl != 0)
			    prtolog("Debug: Opening file", NULL);
			op = 0;
			obuf[0] = 0;
			f1 = genfil(f1, filstr, path, name, outmod);
			if (f1 == NULL) {
			    prtolog("Can't open file", NULL);
			    if (mailstr[0] != 0) {
				/*  Mail error to admin  */
				snprintf(mestr, 511, \
				  "Can't open file %s", filstr);
				mailerr(mestr, mailstr);
			    }
			    if (altpath[0] == 0) {
				resetty();
				if(logfil != NULL)
				    fclose(logfil);
				exit(1);
			    } else {
				/*  Force to raw  */
				f1 = genfil(f1, filstr, altpath, name, \
				  MENRD_RAW);
				if (f1 == NULL) {
				    prtolog("Can't open file", NULL);
				    resetty();
				    if(logfil != NULL)
					fclose(logfil);
				    exit(1);
				}
			    }
			}
			if(debfl != 0)
			    prtolog("Debug: Opened file", filstr);
			tot = 0;	/*  New file  */
			/*  We have a open file and a filename  */
		    } else {
			prtolog("File should be closed", NULL);
		    }
		} else if (buf[i] == 0x03 || buf[i] == 0x1A) {
		    /*  ETX or EOF  */
		    if (buf[i] == 0x03) {
			/*  ETX, send ACK  */
		    	sendch = 0x06;
		    }
		    if (f1 == NULL) {
			prtolog("File should be open", NULL);
		    } else {
			if (obuf[0] != 0) {
			    /*  Flush  */
			    //obuf[op + 1] = 0;
			    obuf[op] = 0;	/*  op points to current char  */
			    datstr[0] = 0;	/*  Clear date string  */
			    /*  File name buffer lengte mode imgpath */
			    tot = tot + procout(f1, filstr, obuf, op, outmod, \
			      imgpath);
			    if (debfl != 0) {
				snprintf(logstr, 511, \
				  "Debug: Wrote %d bytes to procout()", op);
				prtolog(logstr, NULL);
			    }
			}
			/*  Close file  */
			if(debfl != 0)
			    prtolog("Debug: Closing file", filstr);
			fflush(f1);
			fclose(f1);
			f1 = NULL;	/*  Indicate that file is closed  */
			if (debfl != 0) {
			    snprintf(logstr, 511, \
			      "Debug: Wrote %d bytes to file %s", \
			      tot, filstr);
			    prtolog(logstr, NULL);
			}
		    }
		} else if (buf[i] != 0x04) {
		    /*  No EOT, write to file  */
		    if (f1 != NULL && buf[i] != 0) {
			obuf[op] = buf[i];	/*  op points to current char  */
			if (op > 4080) {
			    /*  This should not happen!  */
			    obuf[op + 1] = 0;	/*  Terminate  */
			    datstr[0] = 0;	/*  Clear date string  */
			    tot = tot + procout(f1, filstr, obuf, op + 1, \
			      MENRD_RAW, imgpath);
			    if (debfl != 0) {
				snprintf(logstr, 511, \
				  "Debug: Wrote %d bytes to procout()", op + 1);
				prtolog(logstr, NULL);
			    }
			    op = 0;
			    obuf[0] = 0;
			} else {
			    op++;	/*  op points to next char  */
			}
		    }
		}
		i++;
	    }	/*  End of while (i < len)  */
	}	/*  End of FD_ISSET(fd, &readfs)  */
    }		/*  End of while (contprog != 0)  */
    
    /*  restore the old port settings  */
    resetty();
    /*  Signals to default  */
    //signal(SIGHUP,  SIG_DFL);
    signal(SIGTERM,  SIG_DFL);
    /*  Close file if open  */
    if (f1 != NULL) {
	fflush(f1);
	fclose(f1);
    }
    #ifndef MENRD_USE_TIMEOUT
     /*  Close pipe  */
     close(pfd[0]);
     close(pfd[1]);
    #endif

    /*    Remove PID file    */
    if (pidf[0] != 0) {
	if(unlink(pidf) != 0)
	    prtolog("Can't remove pidfile:", pidf);
    }

    prtolog("menar-read stopped", NULL);

    if(logfil != NULL)
	fclose(logfil);

    return(0);
}

